/* Bootloader. Nimmt Intel Hex Daten per RS232 (600 Baud) entgegen und beschreibt Flash.
 * Passend fuer ATmega16
 *
 * Quelle: http://www.mikrocontroller.net/articles/AVR_Bootloader_in_C_-_eine_einfache_Anleitung
 * Urheber ursprngliche Version: Mario Grafe
 * Lizenz: CC BY-SA 2.0 DE, Namensnennung - Weitergabe unter gleichen Bedingungen 2.0 Deutschland, http://creativecommons.org/licenses/by-sa/2.0/de/
 * nderungen wurden vorgenommen. 
 *
 * Nutzt die UART library von Peter Fleury (siehe uart.c)
 * Nutzt Code von http://www.nongnu.org/avr-libc/user-manual/group__avr__boot.html
 * 
 * Im makefile wird extra angegeben, dass auch die uart.c compiliert werden soll: "SRC = $(TARGET).c uart.c"
 *
 */

#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/boot.h>
#include <util/delay.h>
#include "uart.h"
 
#define BOOT_UART_BAUD_RATE     600     // Baudrate 
#define START_SIGN              ':'      // Hex-Datei Zeilenstartzeichen 
 
// Zustnde des Bootloader-Programms 
#define BOOT_STATE_EXIT	        0        
#define BOOT_STATE_PARSER       1
 
// Zustnde des Hex-File-Parsers 
#define PARSER_STATE_START      0
#define PARSER_STATE_SIZE       1
#define PARSER_STATE_ADDRESS    2
#define PARSER_STATE_TYPE       3
#define PARSER_STATE_DATA       4
#define PARSER_STATE_CHECKSUM   5
#define PARSER_STATE_ERROR      6
 
void program_page (uint32_t, uint8_t *);
static uint16_t hex2num (const uint8_t *, uint8_t);
static uint16_t hex2num (const uint8_t *, uint8_t);
int main(void);

void program_page (uint32_t page, uint8_t *buf)
{
    uint16_t i;
    uint8_t sreg;
 
    sreg = SREG;
    cli();			// Disable interrupts 
 
    eeprom_busy_wait ();
 
    boot_page_erase (page);		// erase memory
    boot_spm_busy_wait ();		// Wait until the memory is erased. 
 
    for (i=0; i<SPM_PAGESIZE; i+=2)
    {
        // Set up little-endian word. 
        uint16_t w = *buf++;
        w += (*buf++) << 8;
 
        boot_page_fill (page + i, w);
    }
 
    boot_page_write (page);     // Store buffer in flash page.		
    boot_spm_busy_wait();       // Wait until the memory is written.
    boot_rww_enable ();			// Reenable RWW-section again. We need this if we want to jump back to the application after bootloading. 
 
    SREG = sreg;				// Re-enable interrupts (if they were ever enabled). 
}
 
/**
	@brief	Konvertiert Intel Hex ASCII nach Byte
	@param	Zeiger auf ASCII Buffer mit Hex-Werten
			Anzahl der zu konvertierenden Bytes
	@return	Zahl
*/
static uint16_t hex2num (const uint8_t * ascii, uint8_t num)
{
    uint8_t  i;
    uint16_t val = 0;
 
    for (i=0; i<num; i++)
    {
        uint8_t c = ascii[i];
 
        // Hex-Ziffer auf ihren Wert abbilden 
        if (c >= '0' && c <= '9')            c -= '0';  
        else if (c >= 'A' && c <= 'F')       c -= 'A' - 10;
        else if (c >= 'a' && c <= 'f')       c -= 'a' - 10;
 
        val = 16 * val + c;
    }
 
    return val;  
}
 
int main(void)
{
    uint16_t        c = 0, 				// Empfangenes Zeichen + Statuscode 
                    hex_addr = 0,		// Intel-HEX Zieladresse 
                    flash_page = 0,		// Zu schreibende Flash-Page            
                    hex_check = 0,		// Intel-HEX Checksumme zum berprfen des Daten 
                    flash_cnt = 0;		// Positions zum Schreiben in der Datenpuffer 
                    
    uint8_t         temp,								// temporre Variable 
                    boot_state = BOOT_STATE_EXIT,		// Flag zum steuern des Programmiermodus 
                    parser_state = PARSER_STATE_START,	// Empfangszustandssteuerung 
                    flash_page_flag = 1,				// Flag zum ermitteln einer neuen Flash-Page 
                    flash_data[SPM_PAGESIZE], 			// Datenpuffer fr die Hexdaten
                    hex_cnt = 0, 						// Position zum Schreiben in den HEX-Puffer 
                    hex_buffer[5], 						// Puffer fr die Umwandlung der ASCII in Binrdaten 
                    hex_size = 0,						// Intel-HEX Datenlnge 
                    hex_data_cnt = 0, 					// Zhler fr die empfangenen HEX-Daten einer Zeile 
                    hex_type = 0, 						// Intel-HEX Recordtype 
                    hex_checksum=0;						// empfangene HEX-Checksumme 

	TCCR0=0;				// Timer Prescaler = STOP (Wichtig!)

	// Fllen der Puffer mit definierten Werten 
    memset(hex_buffer, 0x00, sizeof(hex_buffer));
    memset(flash_data, 0xFF, sizeof(flash_data));
 
    // Interrupt Vektoren verbiegen 
    temp = GICR;
    GICR = temp | (1<<IVCE);
    GICR = temp | (1<<IVSEL);
 
    // Einstellen der Baudrate und aktivieren der Interrupts 
    uart_init( UART_BAUD_SELECT(BOOT_UART_BAUD_RATE,F_CPU) ); 
    sei();
 
    uart_puts("Bootloader. p=programmieren, q=quit\r\n");
    _delay_ms(2000);
 
    do
    {
        c = uart_getc();
        if( !(c & UART_NO_DATA) )
        {
             // Programmzustand: Parser 
             if(boot_state == BOOT_STATE_PARSER)
             {
                  switch(parser_state)
                  {
                      // Warte auf Zeilen-Startzeichen 
                      case PARSER_STATE_START:			
                          if((uint8_t)c == START_SIGN) 
                          {
                              //uart_putc(XOFF);
                              parser_state = PARSER_STATE_SIZE;
                              hex_cnt = 0;
                              hex_check = 0;
                          }
                          break;
                      // Parse Datengre 
                      case PARSER_STATE_SIZE:	
                          hex_buffer[hex_cnt++] = (uint8_t)c;
                          if(hex_cnt == 2)
                          {
                              parser_state = PARSER_STATE_ADDRESS;
                              hex_cnt = 0;
                              hex_size = (uint8_t)hex2num(hex_buffer, 2);
                              hex_check += hex_size;
                           }
                           break;
                      // Parse Zieladresse 
                      case PARSER_STATE_ADDRESS:
                          hex_buffer[hex_cnt++] = (uint8_t)c;
                          if(hex_cnt == 4)
                          {
                              parser_state = PARSER_STATE_TYPE;
                              hex_cnt = 0;
                              hex_addr = hex2num(hex_buffer, 4);
                              hex_check += (uint8_t) hex_addr;
                              hex_check += (uint8_t) (hex_addr >> 8);
                              if(flash_page_flag) 
                              {
                                  flash_page = hex_addr - hex_addr % SPM_PAGESIZE;
                                  flash_page_flag = 0;
                              }
                           }
                           break;
                      // Parse Zeilentyp 
                      case PARSER_STATE_TYPE:	
                           hex_buffer[hex_cnt++] = (uint8_t)c;
                           if(hex_cnt == 2)
                           {
                               hex_cnt = 0;
                               hex_data_cnt = 0;
                               hex_type = (uint8_t)hex2num(hex_buffer, 2);
                               hex_check += hex_type;
                               switch(hex_type)
                               {
                                   case 0: parser_state = PARSER_STATE_DATA; break;
                                   case 1: parser_state = PARSER_STATE_CHECKSUM; break;
                                   default: parser_state = PARSER_STATE_DATA; break;
                               }
                           }
                           break;
                      // Parse Flash-Daten 
                      case PARSER_STATE_DATA:
                          hex_buffer[hex_cnt++] = (uint8_t)c;
                          if(hex_cnt == 2)
                          {
                              uart_putc('.');
                              hex_cnt = 0;
                              flash_data[flash_cnt] = (uint8_t)hex2num(hex_buffer, 2);
                              hex_check += flash_data[flash_cnt];
                              flash_cnt++;
                              hex_data_cnt++;
                              if(hex_data_cnt == hex_size)
                              {
                                  parser_state = PARSER_STATE_CHECKSUM;
                                  hex_data_cnt=0;
                                  hex_cnt = 0;
                              }
                              // Puffer voll -> schreibe Page 
                              if(flash_cnt == SPM_PAGESIZE)
                              {
                                  uart_puts("P\r\n");
                                  _delay_ms(100);
                                  program_page((uint16_t)flash_page, flash_data);
                                  memset(flash_data, 0xFF, sizeof(flash_data));
                                  flash_cnt = 0;
                                  flash_page_flag = 1;
                              }
                          }
                          break;
                      // Parse Checksumme                              
                      case PARSER_STATE_CHECKSUM:
                          hex_buffer[hex_cnt++] = (uint8_t)c;
                          if(hex_cnt == 2)
                          {
                              hex_checksum = (uint8_t)hex2num(hex_buffer, 2);
                              hex_check += hex_checksum;
                              hex_check &= 0x00FF;
                              // Dateiende -> schreibe Restdaten  
                              if(hex_type == 1)
                              {
                                  uart_puts("P\r\n");
                                  _delay_ms(100);
                                  program_page((uint16_t)flash_page, flash_data);
                                  boot_state = BOOT_STATE_EXIT;
                              }
                              // berprfe Checksumme -> muss '0' sein 
                              if(hex_check == 0) parser_state = PARSER_STATE_START;
                              else parser_state = PARSER_STATE_ERROR;
                          }
                          break;			
                      // Parserfehler (falsche Checksumme) 
                      case PARSER_STATE_ERROR:
                          uart_putc('#');
                          break;			
                      default:
                          break;
                  }
             }
             // Programmzustand: UART Kommunikation                
             else if(boot_state != BOOT_STATE_PARSER)
             {
                 switch((uint8_t)c)
                 {
                     case 'p': 
                         boot_state = BOOT_STATE_PARSER;
                         uart_puts("Programmiere den Flash.\r\n");
                         uart_puts("Hex-Datei senden:\r\n");
                         break;
                     case 'q': 
                         boot_state = BOOT_STATE_EXIT;
                         uart_puts("Verlasse Bootloader.\r\n");
                         break;
                 }
             }
        }
    }
    while(boot_state!=BOOT_STATE_EXIT);
 
    uart_puts("Reset AVR!\r\n");
	_delay_ms(2000);				// wichtig, damit Ausgabe auch abgeschlossen
	
    // Interrupt Vektoren wieder gerade biegen 
    temp = GICR;
    GICR = temp | (1<<IVCE);
    GICR = temp & ~(1<<IVSEL);
 
    asm("jmp 0x0000");      // Sprung Anfang Flash
 
    return 0;
}